@@ -1,6 +1,10 @@
 use crate::ast::*;
 use tract_core::internal::*;
 use tract_core::ops::cnn::deconv::adjustments;
+use tract_core::ops::cnn::PaddingSpec;
+use tract_core::ops::cnn::PoolSpec;
+use tract_core::ops::matmul::MatMulQParams;
+use tract_core::ops::nn::DataFormat;
 use tract_itertools::izip;
 use tract_itertools::Itertools;
 
@@ -14,7 +18,9 @@ pub fn external(
     invocation: &ResolvedInvocation,
 ) -> TractResult<TVec<OutletId>> {
     let type_name = invocation.invocation.generic_type_name.unwrap_or(TypeName::Scalar);
-    let dt = if type_name == TypeName::Scalar {
+    let dt = if let Some(Some(dt)) = invocation.dt_from_quant_file.get(0) {
+        *dt
+    } else if type_name == TypeName::Scalar {
         f32::datum_type()
     } else if type_name == TypeName::Logical {
         bool::datum_type()
@@ -32,7 +38,7 @@ pub fn variable(
 ) -> TractResult<TVec<OutletId>> {
     let shape: TVec<usize> = invocation.named_arg_as(builder, "shape")?;
     let label: String = invocation.named_arg_as(builder, "label")?;
-    let tensor = builder
+    let mut tensor = builder
         .proto_model
         .tensors
         .iter()
@@ -40,6 +46,26 @@ pub fn variable(
         .ok_or_else(|| format_err!("No data for tensor {:?}", label))?
         .1
         .clone();
+    if let Some(Some(dt)) = invocation.dt_from_quant_file.get(0) {
+        if dt.size_of() != tensor.datum_type().size_of() {
+            bail!(
+                "Mismatched tensor type for tensor {}: expected {:?}, got {:?}",
+                label,
+                *dt,
+                tensor.datum_type()
+            );
+        }
+        if *dt != tensor.datum_type() {
+            trace!(
+                "Casting tensor {} from {:?} to {:?} when deserializing",
+                label,
+                tensor.datum_type(),
+                *dt
+            );
+            //FIXME: avoid cast by late-loading tensors ?
+            tensor = tensor.cast_to_dt(*dt)?.into_owned().into_arc_tensor()
+        }
+    }
     if tensor.shape() != &*shape {
         bail!(
             "Wrong shape for tensor: {:?}, tensor file says {:?}, graph files says {:?}",
@@ -71,8 +97,8 @@ pub fn reshape(
         }
     }
     if let Some(pos) = replacement.iter().position(|d| *d == (-1).to_dim()) {
-        let product: TDim = replacement.iter().filter(|d| **d != (-1).to_dim()).maybe_product()?;
-        let product_input: TDim = input_shape[start..][..count].iter().maybe_product()?;
+        let product: TDim = replacement.iter().filter(|d| **d != (-1).to_dim()).product();
+        let product_input: TDim = input_shape[start..][..count].iter().product();
         replacement[pos] = product_input.maybe_div(&product)?.0;
     }
 
@@ -98,7 +124,15 @@ pub fn concat(
     invocation: &ResolvedInvocation,
 ) -> TractResult<TVec<OutletId>> {
     let axis: usize = invocation.named_arg_as(builder, "axis")?;
-    let values: TVec<OutletId> = invocation.named_arg_as(builder, "values")?;
+    let mut values: TVec<OutletId> = invocation.named_arg_as(builder, "values")?;
+    if let Some(Some(dt)) = invocation.dt_from_quant_file.get(0) {
+        for value in &mut values {
+            if builder.model.node(value.node).outputs[value.slot].fact.datum_type != *dt {
+                *value = builder.wire(ops::cast::cast(*dt), &[*value])?[0];
+            }
+        }
+    }
+
     builder.wire(ops::array::TypedConcat::concat_vars(axis, values.len()), &values)
 }
 
@@ -220,31 +254,18 @@ pub fn deconv(
     conv_or_deconv(builder, invocation, true)
 }
 
-pub fn conv_or_deconv(
+pub fn read_conv_parameters(
     builder: &mut ModelBuilder,
     invocation: &ResolvedInvocation,
-    deconv: bool,
-) -> TractResult<TVec<OutletId>> {
-    use ops::cnn::deconv::DeconvUnary;
-    use ops::cnn::{ConvUnary, KernelFormat};
-    use ops::cnn::{PaddingSpec, PoolSpec};
-    use ops::nn::DataFormat;
-    let input: OutletId = invocation.named_arg_as(builder, "input")?;
-    let kernel: Arc<Tensor> = invocation.named_arg_as(builder, "filter")?;
-    let input_fact = builder.model.outlet_fact(input)?.clone();
-    if input_fact.rank() != kernel.rank() {
-        bail!(
-            "Convolution input expected as NCHW, filter as OIHW. Got {:?} and {:?}.",
-            input_fact,
-            kernel
-        );
-    }
+    kernel_shape: &[usize],
+    input_fact: &TypedFact,
+) -> TractResult<(usize, PoolSpec)> {
     let mut group = invocation.named_arg_as(builder, "groups")?;
     if group == 0 {
-        group = kernel.shape()[0]
+        group = kernel_shape[0]
     }
-    if input_fact.shape[1] != kernel.shape()[1].to_dim() * group {
-        bail!("Convolution input and kernel channels (second axis in both) must match. Got {:?} and {:?}.", input_fact, kernel);
+    if input_fact.shape[1] != kernel_shape[1].to_dim() * group {
+        bail!("Convolution input and kernel channels (second axis in both) must match. Got {:?} and {:?}.", input_fact, kernel_shape);
     }
     let dilation: TVec<usize> = invocation.named_arg_as(builder, "dilation")?;
     if dilation.len() != 0 && dilation.len() != input_fact.rank() - 2 {
@@ -268,18 +289,69 @@ pub fn conv_or_deconv(
     };
     let pool_spec = PoolSpec::new(
         DataFormat::NCHW,
-        kernel.shape()[2..].into(),
+        kernel_shape[2..].into(),
         padding,
         if dilation.len() > 0 { Some(dilation) } else { None },
         if stride.len() > 0 { Some(stride) } else { None },
-        Some(kernel.shape()[0]),
+        Some(kernel_shape[0]),
     );
-    let bias: Arc<Tensor> = invocation.named_arg_as(builder, "bias")?;
-    let bias: Option<Arc<Tensor>> =
-        if bias.is_uniform() && bias.cast_to_scalar::<f32>()? == 0.0 { None } else { Some(bias) };
 
     let border: String = invocation.named_arg_as(builder, "border")?;
     assert_eq!(border, "constant");
+
+    Ok((group, pool_spec))
+}
+
+pub fn conv_or_deconv(
+    builder: &mut ModelBuilder,
+    invocation: &ResolvedInvocation,
+    deconv: bool,
+) -> TractResult<TVec<OutletId>> {
+    use ops::cnn::deconv::DeconvUnary;
+    use ops::cnn::{ConvUnary, KernelFormat};
+
+    let input: OutletId = invocation.named_arg_as(builder, "input")?;
+    let kernel: Arc<Tensor> = invocation.named_arg_as(builder, "filter")?;
+    let input_fact = builder.model.outlet_fact(input)?.clone();
+    if input_fact.rank() != kernel.rank() {
+        bail!(
+            "Convolution input expected as NCHW, filter as OIHW. Got {:?} and {:?}.",
+            input_fact,
+            kernel
+        );
+    }
+
+    let (group, pool_spec) =
+        read_conv_parameters(builder, invocation, kernel.shape(), &input_fact)?;
+
+    let output_dt =
+        invocation.dt_from_quant_file.get(0).cloned().flatten().unwrap_or(DatumType::F32);
+    let quantized = input_fact.datum_type.is_quantized()
+        || kernel.datum_type().is_quantized()
+        || output_dt.is_quantized();
+    let (b0, b_scale) = input_fact.datum_type.zp_scale();
+    let (a0, a_scale) = kernel.datum_type().zp_scale();
+    let (c0, c_scale) = output_dt.qparams().map(|q| q.zp_scale()).unwrap_or((b0, b_scale));
+    let qparams = if quantized {
+        Some((
+            output_dt,
+            MatMulQParams {
+                a0: AttrOrInput::Attr(Arc::new(a0.into())),
+                a_scale: AttrOrInput::Attr(Arc::new(a_scale.into())),
+                b0: AttrOrInput::Attr(Arc::new(b0.into())),
+                b_scale: AttrOrInput::Attr(Arc::new(b_scale.into())),
+                c0: AttrOrInput::Attr(Arc::new(c0.into())),
+                c_scale: AttrOrInput::Attr(Arc::new(c_scale.into())),
+            },
+        ))
+    } else {
+        None
+    };
+    let bias: Arc<Tensor> = invocation.named_arg_as(builder, "bias")?;
+
+    let bias: Option<Arc<Tensor>> =
+        if bias.is_uniform() && bias.cast_to_scalar::<f32>()? == 0.0 { None } else { Some(bias) };
+
     let op: Box<dyn TypedOp> = if deconv {
         let output_shape = invocation.named_arg_as::<TVec<usize>>(builder, "output_shape")?;
         let output_shape = Some(output_shape).filter(|os| os.len() == pool_spec.rank());
@@ -301,7 +373,14 @@ pub fn conv_or_deconv(
             group,
         ))
     } else {
-        Box::new(ConvUnary::new(pool_spec, KernelFormat::OIHW, kernel.clone(), group, bias, None))
+        Box::new(ConvUnary::new(
+            pool_spec,
+            KernelFormat::OIHW,
+            kernel.clone(),
+            group,
+            bias,
+            qparams,
+        ))
     };
     builder.wire(op, &[input])
 }
@@ -311,8 +390,6 @@ fn pool_spec_for_pools(
     invocation: &ResolvedInvocation,
     shape: &[usize],
 ) -> TractResult<ops::cnn::PoolSpec> {
-    use ops::cnn::{PaddingSpec, PoolSpec};
-    use ops::nn::DataFormat;
     let dilation: TVec<usize> = invocation.named_arg_as(builder, "dilation")?;
     if dilation.len() > 0 && (dilation.len() != shape.len() || dilation[0] != 1 || dilation[1] != 1)
     {
@@ -365,7 +442,8 @@ pub fn max_pool_with_index(
                 );
     }
     let border: String = invocation.named_arg_as(builder, "border")?;
-    assert_eq!(border, "ignore");
+    assert!(&*border == "ignore" || &*border == "constant");
+    //FIXME : constant is not actually supported, but it should be the same in most cases
     let pool_spec = pool_spec_for_pools(builder, invocation, &size)?;
     let op = ops::cnn::MaxPool { pool_spec, with_index_outputs: Some(i64::datum_type()) };
     builder.wire(op, &[input])
@@ -392,7 +470,7 @@ pub fn sum_pool(
                 );
     }
     let border: String = invocation.named_arg_as(builder, "border")?;
-    assert_eq!(border, "ignore");
+    assert!(&*border == "ignore" || &*border == "constant");
     let pool_spec = pool_spec_for_pools(builder, invocation, &size)?;
     let op = ops::cnn::SumPool {
         pool_spec,
@@ -429,7 +507,7 @@ pub fn reduce(
 
     let fact = builder.model.outlet_fact(wire[0])?;
     let input_shape = &builder.model.outlet_fact(input)?.shape;
-    let cardinality: TDim = axes.iter().map(|ax| &input_shape[*ax]).maybe_product()?;
+    let cardinality: TDim = axes.iter().map(|ax| &input_shape[*ax]).product();
     if let Ok(c) = cardinality.to_isize() {
         if fact.datum_type.is_float() {
             let cardinality = tensor0((c as f64).recip())
@@ -449,10 +527,44 @@ pub fn matmul(
     builder: &mut ModelBuilder,
     invocation: &ResolvedInvocation,
 ) -> TractResult<TVec<OutletId>> {
-    let a = invocation.named_arg_as(builder, "A")?;
-    let b = invocation.named_arg_as(builder, "B")?;
+    let a: OutletId = invocation.named_arg_as(builder, "A")?;
+    let b: OutletId = invocation.named_arg_as(builder, "B")?;
     let a_trans = invocation.named_arg_as(builder, "transposeA")?;
     let b_trans = invocation.named_arg_as(builder, "transposeB")?;
+    if let Some(Some(dt)) = &invocation.dt_from_quant_file.get(0) {
+        if let Some(qparams) = dt.qparams() {
+            let (a0, a_scale) =
+                builder.model.node(a.node).outputs[a.slot].fact.datum_type.zp_scale();
+            let (b0, b_scale) =
+                builder.model.node(b.node).outputs[b.slot].fact.datum_type.zp_scale();
+
+            let (c0, c_scale) = qparams.zp_scale();
+            //FIXME: bias is not specified in the nnef format, but whether the bias is a bias or an add later changes the quantized behaviour
+            let bias = builder.model.add_const(
+                format!("{}.bias", invocation.invocation.id),
+                Tensor::zero_dt(DatumType::QI8(qparams), &[1])?,
+            )?;
+            builder.model.node(a.node);
+
+            return builder.wire(
+                ops::matmul::QMatMul {
+                    a_trans,
+                    b_trans,
+                    c_trans: false,
+                    output_type: DatumType::QI8(qparams),
+                    params: MatMulQParams {
+                        a0: AttrOrInput::Attr(Arc::new(a0.into())),
+                        a_scale: AttrOrInput::Attr(Arc::new(a_scale.into())),
+                        b0: AttrOrInput::Attr(Arc::new(b0.into())),
+                        b_scale: AttrOrInput::Attr(Arc::new(b_scale.into())),
+                        c0: AttrOrInput::Attr(Arc::new(c0.into())),
+                        c_scale: AttrOrInput::Attr(Arc::new(c_scale.into())),
+                    },
+                },
+                &[a, b, bias],
+            );
+        }
+    }
     builder.wire(ops::matmul::MatMul { a_trans, b_trans, c_trans: false }, &[a, b])
 }
 
